@@ -1,18 +1,51 @@
 import argparse
-import json
-import logging
 import os
-from sys import stdout
-from collections import namedtuple
 from time import sleep, time
-from typing import Iterable, Optional
-from tfe_run_wait.env_default import EnvDefault
+from typing import Iterable, List, Optional
 
 import requests
+from requests import Response
+
+from tfe_run_wait.argument_types import EnvDefault
+from tfe_run_wait.logger import log
 
 _tfe_api_token = os.getenv("TFE_API_TOKEN")
 
-Run = namedtuple("Run", ["run", "ingress_attributes"])
+
+def _post(path: str, params: dict = {}, data: dict = {}):
+    hdrs = {
+        "Authorization": f"Bearer {_tfe_api_token}",
+        "content-type": "application/vnd.api+json",
+    }
+
+    if path.startswith("/api/") or path.startswith("api/"):
+        url = f'https://app.terraform.io/{path.lstrip("/")}'
+    else:
+        url = f"https://app.terraform.io/api/v2/{path}"
+
+    log.debug("post %s, %s, %s", url, params, data)
+    r = requests.post(url, headers=hdrs, params=params, json=data)
+    if r.status_code != 200:
+        raise TFEError(r)
+
+
+class TFEError(Exception):
+    def __init__(self, response: Response):
+        try:
+            self.errors = response.json().get("errors", [])
+        except:
+            self.errors = [{"status": response.status_code, "title": response.text}]
+
+    def __str__(self):
+        return "\n".join(
+            map(lambda e: "{}: {}".format(e.get("status"), e.get("title")), self.errors)
+        )
+
+    def status(self):
+        return self.errors[0].get("status")
+
+    def title(self):
+        return self.errors[0].get("status")
 
 
 def _get(path: str, params: dict = {}) -> Optional[dict]:
@@ -23,17 +56,18 @@ def _get(path: str, params: dict = {}) -> Optional[dict]:
     else:
         url = f"https://app.terraform.io/api/v2/{path}"
 
+    log.debug("get %s, %s", url, params)
     r = requests.get(url, headers=hdrs, params=params)
     if r.status_code == 200:
         return r.json()["data"]
     elif r.status_code == 404:
         return None
     else:
-        raise Exception(f"failed to get {path}, status code {r.status_code}, {r.text}")
+        raise TFEError(r)
 
 
 def _list(path: str, headers: dict = {}, params: dict = {}) -> Iterable[dict]:
-    prms = {"page[size]": 15}
+    prms = {"page[size]": 100}
     hdrs = {"Authorization": f"Bearer {_tfe_api_token}"}
     if headers:
         hdrs.update(headers)
@@ -45,6 +79,7 @@ def _list(path: str, headers: dict = {}, params: dict = {}) -> Iterable[dict]:
     else:
         url = f"https://app.terraform.io/api/v2/{path}"
 
+    log.debug("get %s, %s", url, params)
     r = requests.get(url, headers=hdrs, params=prms)
     while r.status_code == 200:
         response = r.json()
@@ -58,10 +93,10 @@ def _list(path: str, headers: dict = {}, params: dict = {}) -> Iterable[dict]:
             return
 
     if r.status_code not in (200, 404):
-        raise Exception(f"failed to list {path}, status code {r.status_code}, {r.text}")
+        raise TFEError(r)
 
 
-def find_run_for_commit(workspace: dict, url: str, commit_sha: str) -> Optional[Run]:
+def find_run_for_commit(workspace: dict, url: str, commit_sha: str) -> Optional[dict]:
     for run in _list(f'workspaces/{workspace["id"]}/runs'):
         configuration_version_id = (
             run.get("relationships", {})
@@ -79,8 +114,7 @@ def find_run_for_commit(workspace: dict, url: str, commit_sha: str) -> Optional[
                     "commit-sha"
                 )
                 if ia_clone_url == url and ia_commit_sha == commit_sha:
-                    return Run(run, ingress_attributes)
-
+                    return run
     return None
 
 
@@ -100,29 +134,29 @@ def get_plan(run: dict) -> Optional[dict]:
 
 def wait_until(
     workspace: dict,
-    wait_for_status: str,
+    wait_for_status: List[str],
     clone_url: str,
     commit_sha: str,
     maximum_wait_time_in_seconds: int,
-    verbose: bool = False,
 ):
     run_id = None
     workspace_name = workspace["attributes"]["name"]
     now = start_time = time()
     while (now - start_time) < maximum_wait_time_in_seconds:
-        run = find_run_for_commit(workspace, clone_url, commit_sha)
-        if run:
-            if verbose:
-                logging.info("%s\n", json.dumps(run, indent=2))
+        if not run_id:
+            run = find_run_for_commit(workspace, clone_url, commit_sha)
+        else:
+            run = _get(f"/api/v2/runs/{run_id}")
 
-            run_id = run.run["id"]
-            status = run.run.get("attributes").get("status")
-            if status == wait_for_status:
-                logging.info(
-                    "run %s in workspace %s has reached state %s",
+        if run:
+            run_id = run["id"]
+            status = run.get("attributes").get("status")
+            if status in wait_for_status:
+                log.info(
+                    "%s in workspace %s has reached state %s",
                     run_id,
                     workspace_name,
-                    wait_for_status,
+                    status,
                 )
                 return 0
             elif status in (
@@ -131,43 +165,47 @@ def wait_until(
                 "canceled",
                 "force_canceled",
                 "planned_and_finished",
+                "applied"
             ):
-                logging.error(
-                    "run %s in workspace %s has status %s and can no longer reach the desired status of %s",
+                log.error(
+                    "%s in workspace %s has status %s and can no longer reach the desired status of %s",
                     run_id,
                     workspace_name,
                     status,
-                    wait_for_status,
+                    ', '.join(wait_for_status),
                 )
                 return 1
             else:
-                logging.info(
-                    "run %s in workspace %s has status %s, waiting 15s",
+                log.info(
+                    "%s in workspace %s in status %s, waited %ss",
                     run_id,
                     workspace_name,
                     status,
+                    int(time() - start_time),
                 )
-                sleep(15)
+                sleep(10)
         else:
-            logging.info(
-                "waiting for a run in workspace %s for commit %s in %s to appear",
+            log.info(
+                "waiting %ss for a run in workspace %s for commit %s in %s to appear",
+                int(time() - start_time),
                 workspace_name,
                 commit_sha[0:7],
                 clone_url,
             )
-            sleep(15)
+            sleep(10)
         now = time()
 
     if run_id:
-        logging.error(
-            "time out while waiting for a run %s in workspace %s to reach state %s",
+        log.error(
+            "timed out after %ss waiting for %s in workspace %s to reach state %s",
+            int(time() - start_time),
             run_id,
             workspace_name,
-            wait_for_status,
+            ", ".join(wait_for_status),
         )
     else:
-        logging.error(
-            "time out while waiting for a run in workspace %s for commit %s in %s",
+        log.error(
+            "time out while waiting for a run in workspace %s for commit %s in %s to appear",
             run_id,
             workspace_name,
             commit_sha[0:7],
@@ -176,7 +214,24 @@ def wait_until(
     return 1
 
 
-def main():
+def _get_org_and_workspace(parser, args) -> (dict, dict):
+    global _tfe_api_token
+    if args.token:
+        _tfe_api_token = args.token
+
+    org = _get(f"organizations/{args.organization}")
+    if not org:
+        parser.error(f"unknown organization {args.organization}.")
+
+    workspace = _get(f"organizations/{args.organization}/workspaces/{args.workspace}")
+    if not workspace:
+        parser.error(
+            f"workspace {args.workspace} is unknown in organization {args.organization}."
+        )
+    return (org, workspace)
+
+
+def _wait():
     parser = argparse.ArgumentParser(
         description="wait for TFE run to reach specified state."
     )
@@ -198,8 +253,9 @@ def main():
     parser.add_argument(
         "--wait-for-status",
         required=False,
-        default="planned_and_finished",
+        default=["applied", "planned_and_finished"],
         help="wait state to reach",
+        action='append'
     )
     parser.add_argument(
         "--maximum-wait-time",
@@ -208,44 +264,89 @@ def main():
         default=45 * 60,
         help="for state to be reached in minutes, default 45",
     )
-    parser.add_argument(
-        "--verbose",
-        required=False,
-        default=False,
-        action="store_true",
-        help="show verbose output",
-    )
     args = parser.parse_args()
+    org, workspace = _get_org_and_workspace(parser, args)
 
-    global _tfe_api_token
-    _tfe_api_token = args.token
-
-    org = _get(f"organizations/{args.organization}")
-    if not org:
-        parser.error(f"unknown organization {args.organization}.")
-
-    workspace = _get(f"organizations/{args.organization}/workspaces/{args.workspace}")
-    if not workspace:
-        parser.error(
-            f"workspace {args.workspace} is unknown in organization {args.organization}."
-        )
-
+    log.info(
+        f"waiting for run in {args.organization}:{args.workspace} for commit {args.commit_sha[0:7]} in repository {args.clone_url}"
+    )
     exit(
         wait_until(
             workspace,
             args.wait_for_status,
             args.clone_url,
             args.commit_sha,
-            args.maximum_wait_time * 15,
-            args.verbose,
+            args.maximum_wait_time * 60,
         )
     )
 
 
-if __name__ == "__main__":
-    logging.basicConfig(
-        stream=stdout,
-        level=os.getenv("LOG_LEVEL", "INFO"),
-        format="%(levelname)s: %(message)s",
+def _apply():
+    parser = argparse.ArgumentParser(description="a planned TFE run.")
+    parser.add_argument(
+        "--token",
+        action=EnvDefault,
+        envvar="TFE_API_TOKEN",
+        help="Terraform Enterprise access token, default from TFE_API_TOKEN",
+    )
+
+    parser.add_argument("--organization", required=True, help="of the workspace")
+    parser.add_argument("--workspace", required=True, help="to inspect runs for")
+    parser.add_argument(
+        "--clone-url", required=True, help="of source repository for the run"
+    )
+    parser.add_argument(
+        "--commit-sha", required=True, help="of commit which initiated the run"
+    )
+    parser.add_argument(
+        "--maximum-wait-time",
+        required=False,
+        type=int,
+        default=45 * 60,
+        help="for state to be reached in minutes, default 45 (min)",
+    )
+    parser.add_argument("--comment", required=True, help="to include with the apply")
+
+    args = parser.parse_args()
+    org, workspace = _get_org_and_workspace(parser, args)
+
+    log.info(
+        f"apply run in {args.organization}:{args.workspace} for commit {args.commit_sha[0:7]} in repository {args.clone_url}"
+    )
+    run = find_run_for_commit(workspace, args.clone_url, args.commit_sha)
+    if not run:
+        parser.error(f"no such run found.")
+
+    status = run.get("attributes", {}).get("status")
+    if status == "planned_and_finished":
+        log.info(
+            "%s in workspace %s for commit %s has already been applied",
+            run["id"],
+            args.workspace,
+            args.commit_sha[0:7],
+        )
+        exit(0)
+
+    try:
+        log.error(
+            "apply run %s in workspace %s from status %s",
+            run["id"],
+            args.workspace,
+            status,
+        )
+        _post(f"runs/{run['id']}/actions/apply", data={"comment": args.comment})
+    except TFEError as e:
+        log.error(
+            "apply of run %s in workspace %s failed, %s", run["id"], args.workspace, e
+        )
+        exit(1)
+
+    exit(
+        wait_until(
+            workspace,
+            ["applied"],
+            args.clone_url,
+            args.commit_sha,
+            args.maximum_wait_time * 60,
+        )
     )
-    main()
